class Formulas:
    def atomicWeight(weights):
        """
        Input a list of lists of percent composition (in form of 0.x) and mass and outputs amu.
        """

        return sum([i * j for i, j in weights])
    
    def molarity(molarity=False, moles=False, volume=False):
        """
        Inputs two of molarity, moles, and volume and outputs the other.
        """

        if molarity == False:
            try:
                volume = float(volume)
                moles = float(moles)
                return moles / volume
            except:
                raise Exception("Invalid input. Input a numerical value.")
        elif moles == False:
            try:
                volume = float(volume)
                moles = float(moles)
                return moles * volume
            except:
                raise Exception("Invalid input. Input a numerical value.")
        elif volume == False:
            try:
                volume = float(volume)
                moles = float(moles)
                return moles / molarity
            except:
                raise Exception("Invalid input. Input a numerical value.")
        else:
            raise Exception("Leave one parameter blank to solve for the missing parameter.")


from chemistry.conversions import Conversions
from chemistry.formulas import Formulas
from chemistry.periodic import *

# constants
c = 299792458
a = 6.0221408 * (10 ** 23)
e = -1 * 1.60217663 * (10 ** (-1 * 19))
R = 8.314


import json
from pathlib import Path
import re

def name(ele):
    """
    Inputs either an atomic number or symbol to fetch atomic name.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["name"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["name"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def number(ele):
    """
    Inputs atomic symbol to fetch atomic number.
    """

    if type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return int(i["atomicNumber"])
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic symbol.")

def symbol(ele):
    """
    Inputs an atomic number to fetch atomic symbol.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["atomicNumber"].lower() == ele.lower():
                return i["symbol"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number.")

def mass(ele):
    """
    Inputs either an atomic number or symbol to fetch atomic mass.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return float(data[ele - 1]["atomicMass"].split("(")[0])
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return float(i["atomicMass"].split("(")[0])
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def elecConfig(ele):
    """
    Inputs either an atomic number or symbol to fetch electronic configuration.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["electronicConfiguration"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["electronicConfiguration"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def position(ele):
    """
    Inputs either an atomic number or symbol to fetch atomic position (period, group).
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return (data[ele - 1]["period"], data[ele - 1]["group"])
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return (i["period"], i["group"])
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def elecNeg(ele):
    """
    Inputs either an atomic number or symbol to fetch electron negativity.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["electronegativity"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["electronegativity"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def atomRadius(ele):
    """
    Inputs either an atomic number or symbol to fetch atomic radius (pm).
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["atomicRadius"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["atomicRadius"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def ionRadius(ele):
    """
    Inputs either an atomic number or symbol to fetch ion radius (pm).
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        if data[ele - 1]["ionRadius"] == "unknown":
            return data[ele - 1]["ionRadius"]
        else:
            return int(data[ele - 1]["ionRadius"].split("(")[0])
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                if i["ionRadius"] == "unknown":
                    return i["ionRadius"]
                else:
                    return int(i["ionRadius"].split("(")[0])
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def standardState(ele):
    """
    Inputs either an atomic number or symbol to fetch state of matter at STP.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["standardState"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["standardState"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def ionEnergy(ele):
    """
    Inputs either an atomic number or symbol to fetch ionization energy (kJ/mol).
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["ionizationEnergy"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["ionizationEnergy"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def elecAffinity(ele):
    """
    Inputs either an atomic number or symbol to fetch electron affinity (kJ/mol).
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["electronAffinity"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["electronAffinity"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def meltingPoint(ele):
    """
    Inputs either an atomic number or symbol to fetch melting point (K).
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["meltingPoint"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["meltingPoint"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def boilingPoint(ele):
    """
    Inputs either an atomic number or symbol to fetch boiling point (K).
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["boilingPoint"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["boilingPoint"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def block(ele):
    """
    Inputs either an atomic number or symbol to fetch the block.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["block"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["block"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def year(ele):
    """
    Inputs either an atomic number or symbol to fetch the year discovered.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["yearDiscovered"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["yearDiscovered"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def groupBlock(ele):
    """
    Inputs either an atomic number or symbol to fetch the type of element by group.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        return data[ele - 1]["groupBlock"]
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                return i["groupBlock"]
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def oxidationStates(ele):
    """
    Inputs either an atomic number or symbol to fetch the type of element by group.
    """

    if type(ele) is int:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        if data[ele - 1]["oxidationStates"] == "unknown":
            return data[ele - 1]["oxidationStates"]
        else:
            return list(map(int, data[ele - 1]["oxidationStates"].split(", ")))
    elif type(ele) is str:
        folder = Path(__file__).parent
        with (folder / 'periodic.json').open('r') as f:
            data = json.load(f)
        for i in data:
            if i["symbol"].lower() == ele.lower():
                if i["groupBlock"] == "unknown":
                    return i["oxidationStates"]
                else:
                    return list(map(int, i["oxidationStates"].split(", ")))
        raise Exception("Inputted value is not an atomic number or symbol.")
    else:
        raise Exception("Inputted value is not an atomic number or symbol.")

def molar_mass(molecule):
    """
    Inputs a molecule in type string and returns molar mass of that molecule.
    """

    atoms = []
    current_atom = ""
    current_sub = ""
    for i in range(len(molecule)):
        try:
            int(molecule[i])
            if current_atom != "":
                atoms.append(current_atom)
                current_atom = ""
            current_sub += str(molecule[i])
            if len(molecule) - 1 != i:
                try:
                    int(molecule[i + 1])
                    pass
                except:
                    atoms.append(current_sub)
                    current_sub = ""
        except:
            try:
                str(molecule[i])
                if molecule[i].isupper() == True and current_atom == "":
                    current_atom += molecule[i]
                elif molecule[i].islower() == True:
                    current_atom += molecule[i]
                elif molecule[i].isupper() == True and current_atom != "":
                    atoms.append(current_atom)
                    current_atom = molecule[i]
            except:
                raise Exception("Invalid input. Molecule should only contain letters and numbers.")

    if current_atom != "":
        atoms.append(current_atom)
    if current_sub != "":
        atoms.append(current_sub)
    
    sum = 0
    cur_atom = ""
    for i in range(len(atoms)):
        try:
            int(atoms[i])
            sum += mass(cur_atom) * int(atoms[i])
            cur_atom = ""
        except:
            if len(atoms) - 1 != i:
                if cur_atom != "":
                    sum += mass(cur_atom)
                cur_atom = atoms[i]
            else:
                if cur_atom != "":
                    sum += mass(cur_atom)
                sum += mass(atoms[i])
    return sum


class Conversions:
    global metric_pre
    metric_pre = {"Y": 24, "Z": 21, "E": 18, "P": 15, "T": 12, "G": 9, "M": 6, "k": 3, "h": 2, "da": 1, "base": 0, "d": -1, "c": -2, "m": -3, "μ": -6, "mu": -6, "n": -9, "p": -12, "f": -15, "a": -18, "z": -21, "y": -24}
    
    def k_to_c(k):
        """
        Inputs Kelvin and returns Celcius.
        """

        try:
            return float(k) - 273.15
        except:
            raise Exception("Value inputted is not a number.")
    
    def k_to_f(k):
        """
        Inputs Kelvin and returns Fahrenheit.
        """
        
        try:
            return (float(k) - 273.15) * 9/5 + 32
        except:
            raise Exception("Value inputted is not a number.")
    
    def c_to_k(c):
        """
        Inputs Celvin and returns Kelvin.
        """

        try:
            return float(c) + 273.15
        except:
            raise Exception("Value inputted is not a number.")
    
    def c_to_f(c):
        """
        Inputs Celvin and returns Fahrenheit.
        """

        try:
            return float(c) * 9/5 + 32
        except:
            raise Exception("Value inputted is not a number.")
    
    def f_to_k(f):
        """
        Inputs Fahrenheit and returns Kelvin.
        """

        try:
            round(((float(f) - 32) * 5/9 + 273.15), 5)
        except:
            raise Exception("Value inputted is not a number.")
    
    def f_to_c(f):
        """
        Inputs Fahrenheit and returns Celcius.
        """

        try:
            return round(((float(f) - 32) * 5/9), 5)
        except:
            raise Exception("Value inputted is not a number.")
    
    def metric(num, pre1, pre2):
        """
        Inputs a number, initial metric prefix, wanted metric prefix. Returns number with correct metric prefix.
        """

        try:
            num = float(num)
            if pre1 and pre2 in metric_pre:
                return num * (10 ** (metric_pre[pre1] - metric_pre[pre2]))
            else:
                raise Exception("Inputted prefixes are invalid. Reference [] for valid prefixes.")
        except:
            raise Exception("Value inputted is not a number.")